import time
import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
from utils import START, STOP

class RNNSeqModel(nn.Module):
    """
    Wraps an RNN like model, and add utilities for modeling discrete sequences consisting
    of a fixed set of symbols
    For simplicity, this class assumes batch_size=1

    A sequence model maintains a set of symbols, and models the conditional probability
    p(x_t | h_t) with h_t = forward(x_t, h_{t-1})
    forward_sample draws x_t ~ p(x_t | h_t), with an optional prefix string.
    """
    def __init__(self, rnn, symbols, hidden_size, n_layers=1):
        super(RNNSeqModel, self).__init__()
        self.symbols = symbols
        input_size = len(symbols)
        self.input_size = input_size

        self.hidden_size = hidden_size
        self.n_layers = n_layers
        self.embed_layer = nn.Embedding(input_size, input_size)
        self.rnn = rnn(input_size, hidden_size, n_layers)

        self.output_layer = nn.Linear(hidden_size, input_size)

    def forward(self, input, hidden):
        embed = self.embed_layer(input)
        rnn_output, hidden = self.rnn(embed.view(-1, 1, self.input_size), hidden)
        output = self.output_layer(rnn_output)
        return output, hidden

    def h0(self):
        return Variable(torch.zeros(self.n_layers, 1, self.hidden_size))

    def onehot(self, symbol):
        tensor = torch.zeros(1).long()
        tensor[0] = self.symbols.index(symbol)
        return Variable(tensor)

    def forward_sample(self, prefix=[START], l=300):
        hidden = self.h0()
        predicted = list(prefix)
        # **(5 pts)** Update states to take the prefix into account
        # BEGIN_YOUR_CODE (around 3 lines of code expected)
        raise Exception("Not implemented yet")
        # END_YOUR_CODE
        x = self.onehot(prefix[-1])

        for t in range(l):
            output, hidden = self.forward(x.view(1,-1), hidden)
            output_dist = output.data.view(-1).exp()
            sample = torch.multinomial(output_dist, 1)[0]
            predicted_t = self.symbols[sample]
            predicted += [predicted_t]
            if (predicted_t == STOP):
                return predicted
            x = self.onehot(predicted_t)
        return predicted + ['<exceeded %d>' % l]

    def log_probs(self, xs):
        hidden = self.h0()
        lossf = nn.CrossEntropyLoss()
        logprobs = []
        for t in range(len(xs)-1):
            xt = self.onehot(xs[t])
            xtp1 = self.onehot(xs[t+1])
            output, hidden = self.forward(xt.view(-1, 1), hidden)
            loss = lossf(output.view(1,-1), xtp1)
            logprobs += [-loss.data[0]]
        return logprobs


class RawRNN(nn.Module):
    """
    A slightly (around 30%) slower RNN in basic torch as opposed to C code
    For this assignment, batch_size=1, and n_layers=1
    Recall that for an RNN
           h_t = tanh(V x_t  + W h_(t-1) + b)

    Inputs:
        - input (seq_len, batch_size, input_size)
        - hidden (n_layers, batch_size, hidden_size)

    Outputs:
        - output (seq_len, batch_size, hidden_size): tensor
          containing the hidden units of the last layer of the RNN
          for all steps up to seq_len

        - hidden (num_layers, batch, hidden_size): tensor
          containing the hidden state at the last time step
    """
    def __init__(self, input_size, hidden_size, n_layers=1):
        super(RawRNN, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.n_layers = n_layers
        self.V = nn.Linear(input_size, hidden_size)
        self.W = nn.Linear(hidden_size, hidden_size)
        self.act = nn.Tanh()

    def forward(self, input, hidden):
        seq_length = input.size()[0]
        hiddens = []
        # **(5 pts)** compute hidden states, and put them together using torch.cat
        # BEGIN_YOUR_CODE (around 5 lines of code expected)
        raise Exception("Not implemented yet")
        # END_YOUR_CODE
        return output, hidden

def train_step(model, seq):
    hidden = model.h0()
    criterion = nn.CrossEntropyLoss()
    loss = Variable(torch.zeros(1))
    for t in range(len(seq)-1):
        x = model.onehot(seq[t])
        y = model.onehot(seq[t+1])
        output, hidden = model(x.view(1, -1), hidden)
        loss += criterion(output.view(1, -1), y)
    return loss / (len(seq)-1)

def train_step_vectorized(model, seq):
    hidden = model.h0()
    criterion = nn.CrossEntropyLoss()
    # **(5 pts)** vectorized training code to give a speedup
    # there should not be a loop in here
    # BEGIN_YOUR_CODE (around 10 lines of code expected)
    raise Exception("Not implemented yet")
    # END_YOUR_CODE
    return loss

def evaluate(model, valid):
    perps = []
    for xs in valid:
        perps += [-np.mean(model.log_probs(xs))]
    return np.mean(perps)

def train(model, data, vectorized=False, total_steps=1000, num_reports=10):
    validset = data.validation(100)
    report_interval = max(1, total_steps // num_reports)

    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    start = time.time()
    all_losses = []
    loss_avg = 0
    print('time \tprogress \ttrain_nll \tvalid_nll')
    for step in range(1, total_steps + 1):
        x = data.random()
        model.zero_grad()

        if vectorized:
            loss = train_step_vectorized(model, x)
        else:
            loss = train_step(model, x)

        loss.backward()
        optimizer.step()
        loss_avg += loss.data[0]

        if step % report_interval == 0:
            valid_loss = evaluate(model, validset)
            print('%.2fs \t(%d %d%%) \t%.4f \t%.4f'\
                  % (time.time() - start, step, step / total_steps * 100,\
                     loss_avg / report_interval, valid_loss))
            all_losses.append(valid_loss)
            loss_avg = 0
    return all_losses
